/* 

Copyright 2003-2008 MicroNova (R)
All rights reserved.

Redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following
conditions are met:

    * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

    * Neither the name of MicroNova nor the names of its contributors
    may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

*/

package com.micronova.jsp.tag;

import com.micronova.util.StringUtil;
import java.sql.Connection;
import java.sql.DriverManager;
import java.util.List;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.servlet.jsp.JspTagException;
import javax.servlet.jsp.PageContext;
import javax.servlet.jsp.jstl.core.Config;
import javax.sql.DataSource;
import org.apache.taglibs.standard.tag.common.sql.TransactionTagSupport;

/** stackable transaction tag */

public class TransactionTag extends YuzuTag
{
    protected Object _dataSource;
    protected String _isolation;
    protected int _isolationBefore;

    protected Connection _connection;
    protected boolean _doesOwnConnection;

    public static final String ANCESTORCONNECTION = "com.micronova.jsp.tag.TransactionTag.AncestorConnection";

    public static int fromIsolationSpec(String isolationSpec) throws Exception
    {
        if ("read_committed".equals(isolationSpec))
        {
            return Connection.TRANSACTION_READ_COMMITTED;
        }
        else if ("read_uncommitted".equals(isolationSpec))
        {
            return Connection.TRANSACTION_READ_UNCOMMITTED;
        }
        else if ("repeatable_read".equals(isolationSpec))
        {
            return Connection.TRANSACTION_REPEATABLE_READ;
        }
        else if ("serializable".equals(isolationSpec))
        {
            return Connection.TRANSACTION_SERIALIZABLE;
        }
        else if ("none".equals(isolationSpec))
        {
            return Connection.TRANSACTION_NONE;
        }
        else
        {
            throw new JspTagException("unsupported transaction isolation:" + isolationSpec);
        }
    }

    /** utility to open connection from given datasource spec */

    public static Connection openConnection(Object spec) throws Exception
    {
        Connection connection = null;

        if (spec instanceof DataSource)
        {
            connection = ((DataSource)spec).getConnection();
        }
        else
        {
            String specString = (String)spec;

            if (specString.startsWith("jdbc:"))
            {
                List parts = StringUtil.split(specString, ',');
                
                int partsLength = parts.size();

                String jdbcURL = parts.get(0).toString();
                String driverClassName = parts.get(1).toString();

                String user = null;
                String password = null;

                if (partsLength >= 3)
                {
                    user = parts.get(2).toString();
                }

                if (partsLength >= 4)
                {
                    password = parts.get(3).toString();
                }

                Class.forName(driverClassName, true, Thread.currentThread().getContextClassLoader()).newInstance();           

                if ((user == null) && (password == null))
                {
                    connection = DriverManager.getConnection(jdbcURL);
                }
                else
                {
                    connection = DriverManager.getConnection(jdbcURL, user, password);
                }
            }
            else
            {
                Context context = new InitialContext();
                Context environment = (Context)context.lookup("java:comp/env");

                DataSource dataSource = (DataSource)environment.lookup(specString);

                connection = dataSource.getConnection();
            }
        }

        return connection;
    }

    protected static TransactionTag getAncestorTransactionTag(YuzuTag tag)
    {
        return (TransactionTag)tag.getAncestorTag("com.micronova.jsp.tag.TransactionTag");
    }

    protected static TransactionTagSupport getAncestorTransactionTagSupport(YuzuTag tag)
    {
        return (TransactionTagSupport)tag.getAncestorTag("org.apache.taglibs.standard.tag.common.sql.TransactionTagSupport");
    }

    /** returns given tag's ancestor's connection if possible, otherwise null.  m:transaction takes precedence over sql:transaction */

    public static Connection getAncestorConnection(PageContext pageContext, YuzuTag tag) throws Exception
    {
        TransactionTag transactionTag = getAncestorTransactionTag(tag);
        
        if (transactionTag != null)
        {
            return transactionTag.getConnection();
        }

        TransactionTagSupport transactionTagSupport = getAncestorTransactionTagSupport(tag);

        if (transactionTagSupport != null)
        {
            return transactionTagSupport.getSharedConnection();
        }

        return (Connection)EL.getRequestAttribute(pageContext, ANCESTORCONNECTION);
    }

    protected void init()
    {
        super.init();

        _dataSource = null;
        _isolation = "none";
    }

    protected void prepare() throws Exception
    {
        super.prepare();

        _isolationBefore = fromIsolationSpec(_isolation);
        _connection = null;
        _doesOwnConnection = false;
    }

    protected Connection getConnection(Object spec) throws Exception
    {
        Connection connection = _connection;

        if (connection != null)
        {
            return connection;
        }
        else
        {
            if (spec == null)
            {
                spec = getConfiguration(Config.SQL_DATA_SOURCE, null);
            }

            connection = openConnection(spec);
        }

        return connection;
    }

    protected Connection getConnection() throws Exception
    {
        return getConnection(_dataSource);
    }

    public void initBody() throws Exception
    {
        Connection connection = null;

        Object dataSource = _dataSource;

        if (dataSource == null)
        {
            Connection defaultConnection = (Connection)EL.getRequestAttribute(pageContext, ANCESTORCONNECTION);
            connection = getAncestorConnection(pageContext, this);
        }

        if (connection != null)
        {
            _doesOwnConnection = false;
            _connection = connection;
            _isolationBefore = connection.getTransactionIsolation();
        }
        else
        {
            _doesOwnConnection = true;
            connection = getConnection(dataSource);
            connection.setAutoCommit(false);

            _connection = connection;
        }

        if (connection != null)
        {
            int isolation = fromIsolationSpec(_isolation);

            if (isolation != Connection.TRANSACTION_NONE)
            {
                connection.setTransactionIsolation(isolation);
            }
        }

        super.initBody();
    }
    
    public void afterBody() throws Exception
    {
        super.afterBody();

        if (_doesOwnConnection)
        {
            _connection.commit();
        }
    }

    public void doCatch(Throwable t) throws Throwable
    {
        if (_doesOwnConnection)
        {
            _connection.rollback();
        }

        super.doCatch(t);
    }

    public void doFinally()
    {
        try
        {
            if (_doesOwnConnection)
            {
                try
                {
                    _connection.close();
                }
                catch (Exception ee)
                {
                }
            }
            else
            {
                _connection.setTransactionIsolation(_isolationBefore);
            }
        }
        catch (Exception e)
        {
        }

        super.doFinally();
    }

    public void setDataSource(Object expression) throws Exception
    {
        Object dataSource = evaluateAttribute("dataSource", expression, Object.class, _dataSource);
        
        if (!isEmptyString(dataSource))
        {
            _dataSource = dataSource;
        }
        else
        {
            _dataSource = null;
        }
    }

    public void setIsolation(Object expression) throws Exception
    {
        _isolation = (String)evaluateAttribute("dataSource", expression, String.class, _isolation);
    }
}
