/* 

Copyright 2003-2008 MicroNova (R)
All rights reserved.

Redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following
conditions are met:

    * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

    * Neither the name of MicroNova nor the names of its contributors
    may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

*/

package com.micronova.jsp.tag;

import com.micronova.util.BeanUtil;
import com.micronova.util.IOUtil;
import com.micronova.util.NestedMap;
import com.micronova.util.StringUtil;
import com.micronova.util.TypeUtil;
import com.micronova.util.codec.CodecJSP;
import com.micronova.util.servlet.DispatchHttpServletRequest;
import java.io.Reader;
import java.io.StringWriter;
import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;
import javax.servlet.jsp.JspException;
import javax.servlet.jsp.PageContext;
import javax.servlet.jsp.jstl.core.Config;
import javax.servlet.jsp.tagext.Tag;
import javax.servlet.jsp.tagext.TryCatchFinally;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import org.w3c.dom.Node;

/**

Base class for YUZU tag library.

*/

public class YuzuTag extends YuzuRoot implements TryCatchFinally
{
    /** Object indicating 'default value' */

    public static final String DEFAULT = "default";

    /** name of the call stack in requestScope */

    public static final String CALLSTACK = "com.micronova.jsp.tag.YuzuTag.callStack";

    /** standard callstack fields */

    public static final String TAG = "tag";
    public static final String PAGECONTEXT = "pageContext";
    public static final String HIDDEN = "hidden";

    /** root pageContext in requestScope, if any */

    public static final String ROOTPAGECONTEXT = "com.micronova.jsp.tag.YuzuTag.rootPageContext";

    /** name of this class (used for parent tag search) */

    public static final String CLASSNAME = "com.micronova.jsp.tag.YuzuTag";

    /** page-scoped variable name for the 'tagvalue' */

    public static final String VALUEVAR = "_";

    /** special target name used for returning a value */

    public static final String RETURNTARGET = "return";

    /** 'value' of this tag (either set by the code or 'value' attribute) */

    protected Object _value;

    /** default value set by 'default' attribute */

    protected Object _defaultValue;

    /** test expression for 'default' attribute */

    protected String _test;

    /** variable name for assignment */

    protected String _var;

    /** target Object for assignment */

    protected Object _target;

    /** attribute name to be set */

    protected String _attribute;

    /** property name to be set */

    protected String _property;

    /** scope name of the variable for assignment */

    protected String _scope;
    
    /** class name to be instantiated */

    protected String _className;

    /** codec used on exporting */

    protected String _exportCodec;

    /** codec used on importing */

    protected String _importCodec;

    /** codecs used on processing */

    protected String _processCodec;

    /** codec used on preparation */

    protected String _prepareCodec;

    /** codec used on cleanup */

    protected String _cleanupCodec;

    /** EL expression specifing value to be assigned */

    protected String _assign;

    /** codec used on assignment */

    protected String _assignCodec;

    /** EL expression specifying value to be exported */

    protected String _export;

    /** export mode; 'always' or 'default' */

    protected String _doesExport;

    /** list/map of local variables */

    protected Object _local;

    /** scope of local variables */

    protected String _localScope;

    /** source of tagValue */
    
    protected Object _source;

    /** local variable map */

    protected Map _localMap;

    /** set to true while inside tag body */

    protected boolean _isInBody;

    /** saves current value of VALUEVAR */

    protected Object _valueVarSaved;

    /** set to true if the tag has body */

    protected boolean _hasBody;

    /** set to true if tagValue is assigned */

    protected boolean _isAssigned;

    /** saves current attribute state (for restoration in doEndTag()) */

    protected Map _savedState;

    /** set to true while executing (from doEndTag() to doFinally()) */

    protected boolean _isExecuting;

    /** holds 'current' tagValue */

    protected Object _tagValue;

    /** holds 'default' value */

    protected Object _tagDefaultValue;

    /** set to true if reinitialization in doFinally() is needed */

    protected boolean _doesReinit;

    /** converts scope name to scope int (default = page) */

    public static int fromScopeName(String scopeName) 
    {
        int scope = PageContext.PAGE_SCOPE;

        if ("request".equalsIgnoreCase(scopeName))
        {
            scope = PageContext.REQUEST_SCOPE;
        }
        else if ("session".equalsIgnoreCase(scopeName))
        {
            scope = PageContext.SESSION_SCOPE;
        }
        else if ("application".equalsIgnoreCase(scopeName))
        {
            scope = PageContext.APPLICATION_SCOPE;
        }

        return scope;
    }

    public static void setValueVar(PageContext pageContext, Object value)
    {
        EL.setRequestAttribute(pageContext, VALUEVAR, value);
    }

    public static Object getValueVar(PageContext pageContext)
    {
        return EL.getRequestAttribute(pageContext, VALUEVAR);
    }

    /** obtain variable scope.  Either this tag's scopeName when not null, or any parent's non-null localScope. */

    public int getVariableScope()
    {
        String scopeName = _scope;

        if (!isEmptyString(scopeName))
        {
            return fromScopeName(scopeName);
        }
        else
        {
            List callStack = getCallStack(pageContext);
            
            for (int i = callStack.size(); --i >=0 ;)
            {
                NestedMap entry = (NestedMap)callStack.get(i);
                
                YuzuTag parentTag = (YuzuTag)entry.get(TAG);

                if (parentTag != null)
                {
                    String localScope = parentTag._localScope;
                
                    if (!isEmptyString(localScope))
                    {
                        return fromScopeName(localScope);
                    }
                }
            }
        }

        return fromScopeName(scopeName);
    }

    /** get the value of a scoped variable */

    public static Object getScopedAttribute(PageContext pageContext, String name, int scope)
    {
        return pageContext.getAttribute(name, scope);
    }

    /** obtain configuration named configName, or defaultConfiguration if not found */

    public static Object getConfiguration(PageContext pageContext, String configName, Object defaultConfiguration) throws Exception
    {
        Object configuration = Config.find(pageContext, configName);

        if (configuration instanceof String)
        {
            configuration = EL.evaluateExpression(pageContext, configName, configuration.toString(), Object.class);
        }

        if (configuration == null)
        {
            configuration = defaultConfiguration;
        }

        return configuration;
    }

    public Object getConfiguration(String configName, Object defaultConfiguration) throws Exception
    {
        return getConfiguration(pageContext, configName, defaultConfiguration);
    }

    /** throws a JspException.  If given e is a JspException, it is thrown as-is; otherwise a new JspException with rootCause = e is thrown. */

    public void rethrow(Throwable e) throws JspException
    {
        if (e instanceof JspException)
        {
            throw (JspException)e;
        }
        else
        {
            throw new JspException(e);
        }
    }

    /** constructor */

    public YuzuTag()
    {
        super();

        init();
    }
    
    /** initializes attributes.  Called only once in the constructor. */

    protected void init()
    {
        _value = null;
        _defaultValue = null;
        _test = null;
        _var = null;
        _target = null;
        _property = null;
        _scope = null;
        _attribute = null;
        _className = null;
        _source = null;

        _local = null;
        _localScope = null;
        _export = DEFAULT;
        _assign = DEFAULT;
        _doesExport = DEFAULT;

        _assignCodec = null;
        _exportCodec = null;
        _importCodec = null;
        _processCodec = null;
        _prepareCodec = null;
        _cleanupCodec = null;
    }

    /** initializes household variables.  Always called first in doStartTag(). */

    protected void prepare() throws Exception
    {
        _tagValue = _value;
        _tagDefaultValue = _defaultValue;
        _localMap = null;
        _valueVarSaved = null;
        _isInBody = false;
        _hasBody = false;
        _isAssigned = false;
        _savedState = null;
        _isExecuting = false;
        _doesReinit = false;
    }

    /** called after tag processing is done (in doFinally()) */

    protected void cleanup()
    {
        String cleanupCodec = _cleanupCodec;

        if (cleanupCodec != null)
        {
            try
            {
                applyCodec(cleanupCodec, _tagValue);
            }
            catch (Exception e)
            {
                // exception is ignored
            }
        }
    }

    /** applies codecs to an Object */

    protected Object applyCodec(String codecs, Object object) throws Exception
    {
        return EL.applyCodec(pageContext, codecs, object);
    }

    /*** sets a scoped attribute named 'name' to 'value'.  When 'value' is null, then the attribute is removed.  */

    public static void setScopedAttribute(PageContext pageContext, String name, Object value, int scope)
    {
        if (scope == PageContext.SESSION_SCOPE)
        {
            HttpSession session = ((HttpServletRequest)pageContext.getRequest()).getSession();

            if (value == null)
            {
                session.removeAttribute(name);
            }
            else
            {
                session.setAttribute(name, value);
            }
        }
        else
        {
            if (value == null)
            {
                pageContext.removeAttribute(name, scope);
            }
            else
            {
                pageContext.setAttribute(name, value, scope);
            }
        }
    }

    /** saves 'local' variables.  When 'local' is a map, it also initializes saved variables. */

    protected void saveLocal() throws Exception
    {
        _valueVarSaved = getValueVar(pageContext);

        int localScope = fromScopeName(_localScope);
        Object localObject = _local;
        
        if (localObject instanceof Map)
        {
            Map localMap = new HashMap();

            Iterator iterator = ((Map)localObject).entrySet().iterator();

            while (iterator.hasNext())
            {
                Map.Entry entry = (Map.Entry)iterator.next();

                String key = entry.getKey().toString();
                Object value = entry.getValue();

                localMap.put(key, getScopedAttribute(pageContext, key, localScope));
                setScopedAttribute(pageContext, key, value, localScope);
            }

            _localMap = localMap;
        }
        else if (localObject instanceof List)
        {
            Map localMap = new HashMap();

            Iterator iterator = ((List)localObject).iterator();

            while (iterator.hasNext())
            {
                String key = iterator.next().toString();
                Object value = getScopedAttribute(pageContext, key, localScope);
                localMap.put(key, value);
            }

            _localMap = localMap;
        }
    }

    /** restores local variables */

    protected void restoreLocal()
    {
        try
        {
            setValueVar(pageContext, _valueVarSaved);

            Map localMap = _localMap;

            if (localMap != null)
            {
                int localScope = fromScopeName(_localScope);
            
                Iterator iterator = localMap.entrySet().iterator();

                while (iterator.hasNext())
                {
                    Map.Entry entry = (Map.Entry)iterator.next();

                    setScopedAttribute(pageContext, entry.getKey().toString(), entry.getValue(), localScope);
                }
            }
        }
        catch (Exception e)
        {
        }
    }

    /** instantiate value with given className and optional constructor arguments */

    protected Object instantiateValue(String className) throws Exception
    {
        if (className.indexOf(':') < 0)
        {
            Class c = TypeUtil.forName(className);

            return c.newInstance();
        }
        else
        {
            List partList = StringUtil.split(className, ':');
            className = partList.get(0).toString();

            if ("[]".equals(className))
            {
                Class componentClass = TypeUtil.forName(partList.get(1).toString());
                int argCount = partList.size() - 2;

                int[] dimensions = new int[argCount];

                for (int i = dimensions.length; --i >= 0; )
                {
                    dimensions[i] = TypeUtil.isInteger(evaluateExpression("arrayArgument", partList.get(i + 2).toString(), Integer.class)).intValue();
                }
                
                return Array.newInstance(componentClass, dimensions);
            }
            else
            {
                int argCount = partList.size() - 1;

                Object[] args = new Object[argCount];
                for (int i = 0; i < argCount; i ++)
                {
                    args[i] = evaluateExpression("constructorArgument", partList.get(i + 1).toString(), Object.class);
                }
                
                Class c = TypeUtil.forName(className);
                
                return TypeUtil.newInstance(c, args);
            }
        }
    }

    /** prepares tagValue before tag processing starts.  The returned value is assigned to "_tagValue" when tag evaluation starts. */

    protected Object prepareValue(Object tagValue) throws Exception
    {
        if (tagValue == null)
        {
            String className = _className;
        
            if (className != null)
            {
                tagValue = instantiateValue(className);
            }
        }
        
        String prepareCodec = _prepareCodec;
        
        if (prepareCodec != null)
        {
            tagValue = applyCodec(prepareCodec, tagValue);
        }

        Object source = getSourceObject(_source);

        if (source != null)
        {
            tagValue = copyFromSource(tagValue, source);
        }

        return tagValue;
    }

    /** gets a request-scope stack */

    public static List getStack(PageContext pageContext, String name)
    {
        List list = (List)EL.getRequestAttribute(pageContext, name);

        if (list == null)
        {
            list = new ArrayList();

            NestedMap rootEntry = new NestedMap();
            rootEntry.put(PAGECONTEXT, EL.getRequestAttribute(pageContext, ROOTPAGECONTEXT));

            list.add(rootEntry);

            EL.setRequestAttribute(pageContext, name, list);
        }

        return list;
    }

    /** push to a request-scope stack */

    public static List pushStack(PageContext pageContext, String name, Object object)
    {
        List list = getStack(pageContext, name);
        list.add(object);
        return list;
    }

    /** pop from a request-scope stack */

    public static Object popStack(PageContext pageContext, String name)
    {
        Object object = null;

        List list = getStack(pageContext, name);

        if (!(list.isEmpty()))
        {
            object = list.remove(list.size() - 1);
        }

        return object;
    }

    /** get call stack */

    public static List getCallStack(PageContext pageContext)
    {
        return getStack(pageContext, CALLSTACK);
    }

    /** push call stack */

    public List pushCallStack()
    {
        NestedMap stackEntry = new NestedMap();

        stackEntry.put(TAG, this);
        stackEntry.put(PAGECONTEXT, pageContext);

        return pushStack(pageContext, CALLSTACK, stackEntry);
    }

    /** pull call stack */

    public NestedMap popCallStack()
    {
        return (NestedMap)popStack(pageContext, CALLSTACK);
    }

    /** called before proecessing the body */

    protected void initBody() throws Exception
    {
        _isInBody = true;

        pushCallStack();
    }

    /** called after processing the body */

    protected void afterBody() throws Exception
    {
        popCallStack();

        _isInBody = false;
        _hasBody = true;
    }

    public void doInitBody() throws JspException
    {
        try
        {
            super.doInitBody();

            initBody();
        }
        catch (Exception e)
        {
            rethrow(e);
        }
    }

    public int doAfterBody() throws JspException
    {
        try
        {
            afterBody();

            return super.doAfterBody();
        }
        catch (Exception e)
        {
            rethrow(e);
        }

        // dummy

        return 0;
    }

    public int doStartTag() throws JspException
    {
        try
        {
            _isExecuting = true;

            prepare();

            doPrepare();

            saveLocal();

            setValueVar(pageContext, _tagValue);
        }
        catch (Exception e)
        {
            rethrow(e);
        }

        return EVAL_BODY_BUFFERED;
    }

    public int doEndTag() throws JspException
    {
        try
        {
            Object importedValue = doImport(_tagValue);

            /** apply default */

            Object processValue = doDefault(importedValue, _tagDefaultValue, _test);

            /** process value */

            Object processedValue = doProcess(processValue);

            /** assign value */

            boolean isAssigned = doAssign(processedValue);

            _isAssigned = isAssigned;

            /** export value */

            if (doesExport(processedValue, isAssigned))
            {
                doExport(processedValue);
            }
        }
        catch (Throwable e)
        {
            rethrow(e);
        }

        return EVAL_PAGE;
    }

    /** returns true if tag body needs to be imported.  Default implementation returns true if value is null. */

    protected boolean doesImport(Object value)
    {
        return (value == null);
    }


    /** imports body into an Object applying importCodec if doesImport() returns true, otherwise returns value itself. */

    protected Object importBody(Object value) throws Exception
    {
        if (doesImport(value))
        {
            String bodyString = bodyContent.getString();

            if (!isEmptyString(bodyString))
            {
                String importCodec = _importCodec;

                if (importCodec != null)
                {
                    return applyCodec(importCodec, bodyString);
                }
                else
                {
                    return bodyString;
                }
            }
        }

        return value;
    }

    /** prepares value, called upon tag opening */

    protected void doPrepare() throws Exception
    {
        _tagValue = prepareValue(_tagValue);
    }

    /** apply default logic */

    protected Object doDefault(Object value, Object defaultValue, String test) throws Exception
    {
        if (test == null)
        {
            if (isEmptyString(value))
            {
                value = defaultValue;
            }
        }
        else
        {
            setValueVar(pageContext, value);

            if (!TypeUtil.isTrue(EL.applyCodec(pageContext, test, value)))
            {
                value = defaultValue;
            }
        }

        return value;
    }

    /** does 'import' stage and returns 'importedValue'. */

    protected Object doImport(Object tagValue) throws Exception
    {
        Object importedValue = tagValue;

        if (_hasBody)
        {
            importedValue = importBody(tagValue);
        }

        return importedValue;
    }

    /** If true, then exports the value   Default implementation returns true if isAssigned is false (i.e., "export when not assigned") */
   
    protected boolean doesExport(Object tagValue, boolean isAssigned)
    {
        return (("always".equals(_doesExport)) || (!isAssigned));
    }
    
    /** exports given value */

    protected void exportValue(Object tagValue) throws Exception
    {
        if (tagValue instanceof Node)
        {
            Node node = (Node)tagValue;

            Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");

            StringWriter writer = new StringWriter();

            Source source = new DOMSource(node);
            Result result = new StreamResult(writer);
            
            transformer.transform(source, result);
        
            pageContext.getOut().print(writer.toString());
        }
        else if (tagValue instanceof Reader)
        {
            IOUtil.copy((Reader)tagValue, pageContext.getOut());
        }
        else
        {
            pageContext.getOut().print(tagValue);
        }
    }

    /** returns value to be exported */

    protected Object getExportValue(Object tagValue) throws Exception
    {
        String export = _export;

        if ((export == null) || (export.length() == 0))
        {
            return null;
        }

        if ("_assign".equals(export))
        {
            export = _assign;
        }

        if (export != DEFAULT)
        {
            tagValue = evaluateExpression("export", export, Object.class);
        }

        if (tagValue != null)
        {
            String exportCodec = _exportCodec;

            if (exportCodec != null)
            {
                tagValue = applyCodec(exportCodec, tagValue);
            }
        }

        return tagValue;
    }

    /** does 'Export' stage */

    protected void doExport(Object tagValue) throws Exception
    {
        setValueVar(pageContext, tagValue);

        Object exportObject = getExportValue(tagValue);

        if (exportObject != null)
        {
            exportValue(exportObject);
        }
    }

    /** returns the closest ancestor tag of given class using callStack for yuzu tag descendants */

    protected Object getAncestorTag(String className)
    {
        try
        {
            Class c = Class.forName(className);
            Class yuzuClass = Class.forName(CLASSNAME);

            if (yuzuClass.isAssignableFrom(c))
            {
                List callStack = getCallStack(pageContext);

                for (int i = callStack.size(); --i >=0 ;)
                {
                    NestedMap entry = (NestedMap)callStack.get(i);

                    YuzuTag parent = (YuzuTag)entry.get(TAG);
                    boolean isHidden = TypeUtil.isTrue(entry.get(HIDDEN));

                    if ((c.isInstance(parent)) && (!isHidden))
                    {
                        return parent;
                    }
                }
            }
            else
            {
                Tag parent = this;

                while ((parent = parent.getParent()) != null)
                {
                    if (c.isInstance(parent))
                    {
                        return parent;
                    }
                }
            }
        }
        catch (Exception e)
        {
            // ignored
        }
            
        return null;
    }

    /** tries to set target property when property is not empty.  Returns true if assigned, otherwise false */

    protected boolean setTargetProperty(Object targetObject, String targetProperty, Object assignValue) throws Exception
    {
        if (!isEmptyString(targetProperty))
        {
            BeanUtil.setProperty(targetObject, targetProperty, assignValue);
            return true;
        }
        else
        {
            return false;
        }
    }

    /** does 'Assign' stage */

    protected boolean doAssign(Object assignValue) throws Exception
    {
        setValueVar(pageContext, assignValue);

        String assign = _assign;

        if ("_export".equals(assign))
        {
            assign = _export;
        }

        if (assign != DEFAULT)
        {
            assignValue = evaluateExpression("assign", assign, Object.class);
        }

        String assignCodec = _assignCodec;

        if (assignCodec != null)
        {
            assignValue = applyCodec(assignCodec, assignValue);
        }

        boolean isAssigned = false;

        String attribute = _attribute;

        if (attribute != null)
        {
            Object ancestorTag = getAncestorTag(CLASSNAME);

            if (ancestorTag != null)
            {
                if (setTargetProperty(ancestorTag, attribute, assignValue))
                {
                    isAssigned = true;

                    if (ancestorTag instanceof YuzuTag)
                    {
                        ((YuzuTag)ancestorTag)._doesReinit = true;
                    }
                }

                if ("value".equals(attribute))
                {
                    _valueVarSaved = assignValue;
                }
            }
        }

        Object targetObject = _target;
        String property = _property;

        if (RETURNTARGET.equals(targetObject))
        {
            HttpServletRequest request = (HttpServletRequest)pageContext.getRequest();

            if (request instanceof DispatchHttpServletRequest)
            {
                Object dispatchObject = ((DispatchHttpServletRequest)request).getDispatchObject();
                targetObject = TypeUtil.isNestedMap(dispatchObject);
                property = "@_return.value";

                if (assignValue == null)
                {
                    assignValue = "";
                }
            }
            else
            {
                targetObject = null;
                property = null;
            }
        }

        if (targetObject instanceof String)
        {
            targetObject = evaluateExpression("target", EL.replaceEvalEscape(targetObject.toString()), Object.class);
        }

        String targetProperty = null;

        List assignValueList = null;

        if (property != null)
        {
            targetProperty = property;

            if (targetObject == null)
            {
                targetObject = getAncestorTag(CLASSNAME);

                if (targetObject != null)
                {
                    targetObject = ((YuzuTag)targetObject)._tagValue;
                }
            }

            if (targetObject != null)
            {
                List targetPropertyList = null;

                if (!targetProperty.startsWith("#"))
                {
                    targetPropertyList = StringUtil.split(targetProperty, ',');
                }

                if ((targetPropertyList != null) && (targetPropertyList.size() > 1))
                {
                    Iterator targetPropertyIterator = targetPropertyList.iterator();

                    assignValueList = TypeUtil.isList(assignValue);

                    if (assignValueList != null)
                    {
                        Iterator assignValueIterator = assignValueList.iterator();

                        while (targetPropertyIterator.hasNext())
                        {
                            String subTargetProperty = (String)targetPropertyIterator.next();
                            Object subAssignValue = null;

                            if (assignValueIterator.hasNext())
                            {
                                subAssignValue = assignValueIterator.next();
                            }

                            if (!"".equals(subTargetProperty))
                            {
                                if (setTargetProperty(targetObject, subTargetProperty, subAssignValue))
                                {
                                    isAssigned = true;
                                }
                            }
                        }
                    }
                    else if (assignValue instanceof Map)
                    {
                        Map assignValueMap = (Map)assignValue;

                        while (targetPropertyIterator.hasNext())
                        {
                            String subTargetProperty = (String)targetPropertyIterator.next();

                            if (!("".equals(subTargetProperty)))
                            {
                                if (setTargetProperty(targetObject, subTargetProperty, assignValueMap.get(subTargetProperty)))
                                {
                                    isAssigned = true;
                                }
                            }
                        }
                    }
                    else
                    {
                        while (targetPropertyIterator.hasNext())
                        {
                            String subTargetProperty = (String)targetPropertyIterator.next();

                            if (!("".equals(subTargetProperty)))
                            {
                                if (setTargetProperty(targetObject, subTargetProperty, assignValue))
                                {
                                    isAssigned = true;
                                }
                            }
                        }
                    }
                }
                else
                {
                    if (setTargetProperty(targetObject, targetProperty, assignValue))
                    {
                        isAssigned = true;
                    }
                }
            }
        }

        String var = _var;

        if (!isEmptyString(var))
        {
            int scope = getVariableScope();

            List varList = StringUtil.split(var, ',');

            if (varList.size() > 1)
            {
                Iterator varListIterator = varList.iterator();

                if (assignValueList == null)
                {
                    assignValueList = TypeUtil.isList(assignValue);
                }
                
                if (assignValueList != null)
                {
                    Iterator assignValueIterator = assignValueList.iterator();

                    while (varListIterator.hasNext())
                    {
                        String subVar = (String)varListIterator.next();
                        
                        Object subAssignValue = null;
                        
                        if (assignValueIterator.hasNext())
                        {
                            subAssignValue = assignValueIterator.next();
                        }
                        
                        if (!("".equals(subVar)))
                        {
                            setScopedAttribute(pageContext, subVar, subAssignValue, scope);
                            isAssigned = true;
                        }
                    }
                }
                else if (assignValue instanceof Map)
                {
                    Map assignValueMap = (Map)assignValue;

                    while (varListIterator.hasNext())
                    {
                        String subVar = (String)varListIterator.next();
                        
                        if (!("".equals(subVar)))
                        {
                            setScopedAttribute(pageContext, subVar, assignValueMap.get(subVar), scope);
                            isAssigned = true;
                        }
                    }
                }
                else
                {
                    while (varListIterator.hasNext())
                    {
                        String subVar = (String)varListIterator.next();
                        
                        if (!("".equals(subVar)))
                        {
                            setScopedAttribute(pageContext, subVar, assignValue, scope);
                            isAssigned = true;
                        }
                    }
                }
            }
            else
            {
                setScopedAttribute(pageContext, var, assignValue, scope);
                isAssigned = true;
            }
        } 

        return isAssigned;
    }

    /** processes given value and returns 'processedValue' */

    protected Object processValue(Object tagValue) throws Exception
    {
        return tagValue;
    }

    /* does 'Process' stage */
    
    protected Object doProcess(Object processValue) throws Exception
    {
        Object object = processValue(processValue);
        
        String processCodec = _processCodec;
        
        if (processCodec != null)
        {
            object = applyCodec(processCodec, object);
        }

        return object;
    }

    public void doCatch(Throwable t) throws Throwable
    {
        if (_isInBody)
        {
            popCallStack();
        }

        throw t;
    }

    public void doFinally()
    {
        cleanup();

        restoreLocal();

        _isExecuting = false;

        _isInBody = true;

        Map savedState = _savedState;

        if (savedState != null)
        {
            Iterator iterator = savedState.entrySet().iterator();

            while (iterator.hasNext())
            {
                Map.Entry entry = (Map.Entry)iterator.next();
                String key = (String)entry.getKey();
                Object value = entry.getValue();
                try
                {
                    BeanUtil.setProperty(this, key, value);
                }
                catch (Exception e)
                {
                    e.printStackTrace();
                }
            }
        }

        _savedState = null;

        _isInBody = false;

        if (_doesReinit)
        {
            init();
        }
    }

    /** Returns true if given object is 'empty' (null, "", empty Collection, or empty array */

    public static boolean isEmpty(Object object)
    {
        return TypeUtil.isEmpty(object);
    }

    /** Returns true if given object is an empty string (null or "") */

    public static boolean isEmptyString(Object object)
    {
        return TypeUtil.isEmptyString(object);
    }

    /** Attribute evaluator to be called inside attribute setters.
        To allow nested attribute setters, returns given expression as-is
        if called inside the body (when _isInBody is true).
        If called during execution (between doStartTag() and doEndTag()),
        currentValue is saved for restoration after execution (to keep
        the tag state unchanged after execution for more aggressive tag reuse
        by containers other than tomcat). "name" must be the bean property
        name of the attribute.
    */

    protected Object evaluateAttribute(String name, Object expression, Class valueClass, Object currentValue) throws Exception
    {
        Object value = null;

        if (_isInBody)
        {
            value = expression;
        }
        else if (expression != null)
        {
            value = evaluateAttributeExpression(name, expression, valueClass);
        }

        if (_isExecuting)
        {
            Map savedState = _savedState;
            if (savedState == null)
            {
                savedState = new HashMap();
                _savedState = savedState;
            }

            if (savedState.get(name) == null)
            {
                savedState.put(name, currentValue);
            }
        }

        return value;
    }

    public void setVar(Object expression) throws Exception
    {
        _var = (String)evaluateAttribute("var", expression, String.class, _var);
    }

    public void setScope(Object expression) throws Exception
    {
        _scope = (String)evaluateAttribute("scope", expression, String.class, _scope);
    }

    public void setTarget(Object expression) throws Exception
    {
        _target = evaluateAttribute("target", expression, Object.class, _target);
    }

    public void setProperty(Object expression) throws Exception
    {
        _property = (String)evaluateAttribute("property", expression, String.class, _property);
    }

    public void setValue(Object expression) throws Exception
    {
        _value = evaluateAttribute("value", expression, Object.class, _value);
        _tagValue = _value;
    }

    public void setDefault(Object expression) throws Exception
    {
        _defaultValue = evaluateAttribute("default", expression, Object.class, _defaultValue);
    }

    public void setDefaultValue(Object expression) throws Exception
    {
        setDefault(expression);
    }

    public void setClassName(Object expression) throws Exception
    {
        _className = (String)evaluateAttribute("className", expression, EL.class, _className);
    }

    public void setTest(Object expression) throws Exception
    {
        _test = (String)evaluateAttribute("test", expression, EL.class, _test);
    }

    public void setAttribute(Object expression) throws Exception
    {
        _attribute = (String)evaluateAttribute("attribute", expression, String.class, _attribute);
    }

    public void setLocal(Object expression) throws Exception
    {
        Object local = evaluateAttribute("local", expression, Object.class, _local);

        if (local instanceof String)
        {
            String localString = local.toString();

            if (localString.startsWith("{") || localString.startsWith("<"))
            {
                local = TypeUtil.isNestedMap(localString);
            }
            else 
            {
                List list = TypeUtil.isList(local.toString().split("[ ,\t\r\n]+"));

                if ((list == null) || (list.isEmpty()))
                {
                    local = TypeUtil.isNestedMap(localString);
                }
                else
                {
                    local = list;
                }
            }
        }
        else if (!(local instanceof Map))
        {
            local = TypeUtil.isList(local);
        }

        _local = local;
    }

    public void setLocalScope(Object expression) throws Exception
    {
        _localScope = (String)evaluateAttribute("localScope", expression, String.class, _localScope);
    }

    public void setExport(Object expression) throws Exception
    {
        _export = (String)evaluateAttribute("export", expression, EL.class, _export);
    }

    public void setAssign(Object expression) throws Exception
    {
        _assign = (String)evaluateAttribute("assign", expression, EL.class, _assign);
    }

    public void setExportCodec(Object expression) throws Exception
    {
        _exportCodec = (String)evaluateAttribute("exportCodec", expression, EL.class, _exportCodec);
    }

    public void setAssignCodec(Object expression) throws Exception
    {
        _assignCodec = (String)evaluateAttribute("assignCodec", expression, EL.class, _assignCodec);
    }

    public void setImportCodec(Object expression) throws Exception
    {
        _importCodec = (String)evaluateAttribute("importCodec", expression, EL.class, _importCodec);
    }

    public void setProcessCodec(Object expression) throws Exception
    {
        _processCodec = (String)evaluateAttribute("processCodec", expression, EL.class, _processCodec);
    }

    public void setCodec(Object expression) throws Exception
    {
        setProcessCodec(expression);
    }

    public void setPrepareCodec(Object expression) throws Exception
    {
        _prepareCodec = (String)evaluateAttribute("prepareCodec", expression, EL.class, _prepareCodec);
    }

    public void setCleanupCodec(Object expression) throws Exception
    {
        _cleanupCodec = (String)evaluateAttribute("cleanupCodec", expression, EL.class, _cleanupCodec);
    }

    public void setDoesExport(Object expression) throws Exception
    {
        _doesExport = (String)evaluateAttribute("doesExport", expression, String.class, _doesExport);
    }

    protected Object getSourceObject(Object source) throws Exception
    {
        if (source != null)
        {
            boolean doesEval = false;

            if (source instanceof String)
            {
                String sourceString = (String)source;

                if (sourceString.startsWith("@"))
                {
                    doesEval = true;
                    source = sourceString.substring(1);
                }

                source = TypeUtil.isNestedMap(source);

                if (doesEval)
                {
                    source = CodecJSP.eval(pageContext, source);
                }
            }
        }

        return source;
    }

    protected Object copyFromSource(Object tagValue, Object source) throws Exception
    {
        if ((tagValue != null) && (source != null))
        {
            List list = null;
            NestedMap sourceMap = null;

            if (!(source instanceof NestedMap))
            {
                list = TypeUtil.isList(source);

                if (list == null)
                {
                    if (source instanceof String)
                    {
                        sourceMap = TypeUtil.isNestedMap(source);
                    }
                    else
                    {
                        sourceMap = BeanUtil.getMap(source);
                    }
                }
            }
            else
            {
                sourceMap = (NestedMap)source;
            }

            if (list == null)
            {
                if (sourceMap != null)
                {
                    list = (List)sourceMap.get("_");
                    sourceMap.remove("_");
                }
            }

            if (list != null)
            {
                for (int i = 0; i < list.size(); i ++)
                {
                    BeanUtil.setProperty(tagValue, Integer.toString(i), list.get(i));
                }
            }

            if (sourceMap != null)
            {
                Iterator iterator = sourceMap.entrySet().iterator();

                while (iterator.hasNext())
                {
                    Map.Entry entry = (Map.Entry)iterator.next();
                    BeanUtil.setProperty(tagValue, entry.getKey().toString(), entry.getValue());
                }
            }
        }

        return tagValue;
    }

    public void setSource(Object expression) throws Exception
    {
        Object source = evaluateAttribute("source", expression, Object.class, _source);

        if (_isInBody)
        {
            copyFromSource(_tagValue, getSourceObject(source));
        }
        else
        {
            _source = source;
        }
    }
}
