/* 

Copyright 2003-2008 MicroNova (R)
All rights reserved.

Redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following
conditions are met:

    * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

    * Neither the name of MicroNova nor the names of its contributors
    may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

*/


package com.micronova.util;

import java.io.ByteArrayInputStream;
import java.io.CharArrayReader;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.Types;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

/** ObjectSource for SQL query template rendering */

public class SQLObjectSource implements ObjectSource
{
    /** SQL Type names and index for OUT support */

    public static HashMap TYPES;

    static
    {
        TYPES = new HashMap();

        TYPES.put("ARRAY", new Integer(Types.ARRAY));
        TYPES.put("BIGINT", new Integer(Types.BIGINT));
        TYPES.put("BINARY", new Integer(Types.BINARY));
        TYPES.put("BIT", new Integer(Types.BIT));
        TYPES.put("BLOB", new Integer(Types.BLOB));
        TYPES.put("BOOLEAN", new Integer(Types.BOOLEAN));
        TYPES.put("CHAR", new Integer(Types.CHAR));
        TYPES.put("CLOB", new Integer(Types.CLOB));
        TYPES.put("DATALINK", new Integer(Types.DATALINK));
        TYPES.put("DATE", new Integer(Types.DATE));
        TYPES.put("DECIMAL", new Integer(Types.DECIMAL));
        TYPES.put("DISTINCT", new Integer(Types.DISTINCT));
        TYPES.put("DOUBLE", new Integer(Types.DOUBLE));
        TYPES.put("FLOAT", new Integer(Types.FLOAT));
        TYPES.put("INTEGER", new Integer(Types.INTEGER));
        TYPES.put("JAVA_OBJECT", new Integer(Types.JAVA_OBJECT));
        TYPES.put("LONGVARBINARY", new Integer(Types.LONGVARBINARY));
        TYPES.put("LONGVARCHAR", new Integer(Types.LONGVARCHAR));
        TYPES.put("NULL", new Integer(Types.NULL));
        TYPES.put("NUMERIC", new Integer(Types.NUMERIC));
        TYPES.put("OTHER", new Integer(Types.OTHER));
        TYPES.put("REAL", new Integer(Types.REAL));
        TYPES.put("REF", new Integer(Types.REF));
        TYPES.put("SMALLINT", new Integer(Types.SMALLINT));
        TYPES.put("STRUCT", new Integer(Types.STRUCT));
        TYPES.put("TIME", new Integer(Types.TIME));
        TYPES.put("TIMESTAMP", new Integer(Types.TIMESTAMP));
        TYPES.put("TINYINT", new Integer(Types.TINYINT));
        TYPES.put("VARBINARY", new Integer(Types.VARBINARY));
        TYPES.put("VARCHAR", new Integer(Types.VARCHAR));
    }

    public static final String DEFAULTOUTTYPE = "BIGINT";
    public static final String DEFAULTOUTNAME = "OUTPUT";

    protected List _list;
    protected boolean _isCallable;
    protected ObjectSource _objectSource;

    /** returns true if this is a callable statement */

    public boolean isCallable()
    {
        return _isCallable;
    }

    /** returns a map of output variables (when isCallable) */

    public Map getOutputs(CallableStatement callableStatement) throws Exception
    {
        Map map = new HashMap();

        List list = _list;

        for (int i = 0; i < list.size(); i ++)
        {
            Object[] entry = (Object[])list.get(i);

            Object name = entry[0];
            Integer typeInt = (Integer)entry[1];
            String typeName = (String)entry[2];

            if (typeInt != null)
            {
                map.put(name.toString(), callableStatement.getObject(i + 1));
            }
        }

        return map;
    }

    public SQLObjectSource(ObjectSource objectSource, List list)
    {
        _list = list;
        _objectSource = objectSource;
        _isCallable = false;
    }

    public SQLObjectSource(ObjectSource objectSource)
    {
        this(objectSource, new ArrayList());
    }

    public Object getObject(Object client, Object key)
    {
        try
        {
            String keyString = key.toString();

            if (keyString.indexOf('#') >= 0)
            {
                _isCallable = true;
                String[] specs = key.toString().split("#");

                String inKey = (specs.length > 0) ? specs[0] : "";

                boolean isInOut = !("".equals(inKey));

                String name = (specs.length > 1) ? specs[1] : DEFAULTOUTNAME;
                String typeIntName = (specs.length > 2) ? specs[2] : DEFAULTOUTTYPE;
                Integer typeInt = (Integer)TYPES.get(typeIntName);
                String typeName = (specs.length > 3) ? specs[3] : typeIntName;

                if (isInOut)
                {
                    _list.add(new Object[]{name, typeInt, typeName, _objectSource.getObject(this, inKey)});
                }
                else
                {
                    _list.add(new Object[]{name, typeInt, typeName});
                }
                return "?";
            }
            else
            {
                Object replacement = _objectSource.getObject(this, key);
            
                if (replacement != null)
                {
                    List list = null;

                    if ((!(replacement instanceof byte[])) && (!(replacement instanceof char[])))
                    {
                        list = TypeUtil.isList(replacement);
                    }

                    // expand into "?,?,?,..." if replacement is a List
                    
                    if (list != null)
                    {
                        StringBuffer buffer = new StringBuffer();
                        String separator = "";
                        
                        Iterator iterator = list.iterator();
                        
                        while (iterator.hasNext())
                        {
                            _list.add(new Object[]{iterator.next(), null, null});
                            
                            buffer.append(separator);
                            buffer.append("?");
                            
                            separator = ",";
                        }
                        
                        return buffer.toString();
                    }
                }
                
                _list.add(new Object[]{replacement, null, null});
                return "?";
            }
        }
        catch (Exception e)
        {
            throw new RuntimeException(e);
        }
    }

    public PreparedStatement apply(PreparedStatement preparedStatement) throws Exception
    {
        List list = _list;

        for (int i = 0; i < list.size(); i ++)
        {
            Object[] entry = (Object[])list.get(i);

            Object replacement = entry[0];
            Integer typeInt = (Integer)entry[1];
            String typeName = (String)entry[2];

            boolean hasReplacement = true;

            if (typeInt != null)
            {
                ((CallableStatement)preparedStatement).registerOutParameter(i + 1, typeInt.intValue(), typeName);

                if (entry.length <= 3)
                {
                    hasReplacement = false;
                }
                else
                {
                    replacement = entry[3];
                }
            }

            if (hasReplacement)
            {
                int j = i + 1;

                if (replacement != null)
                {
                    if (replacement instanceof java.util.Date)
                    {
                        replacement = new java.sql.Timestamp(((java.util.Date)replacement).getTime());
                    }

                    if (replacement instanceof byte[])
                    {
                        byte[] bytes = (byte[])replacement;
                        preparedStatement.setBinaryStream(j, new ByteArrayInputStream(bytes), bytes.length);
                    }
                    else if (replacement instanceof char[])
                    {
                        char[] chars = (char[])replacement;
                        preparedStatement.setCharacterStream(j, new CharArrayReader(chars), chars.length);
                    }
                    else
                    {
                        preparedStatement.setObject(j, replacement);
                    }
                }
                else
                {
                    preparedStatement.setNull(j, Types.CHAR);
                }
            }
        }
        
        return preparedStatement;
    }

    public PreparedStatement render(Connection connection, String query, Pattern pattern, boolean isRecursive, boolean doesAllowGroup) throws Exception
    {
        String renderedQuery = (String)Template.render(query, pattern, 1, this, this, isRecursive, doesAllowGroup);

        if (renderedQuery.startsWith("{"))
        {
            _isCallable = true;
        }

        if (_isCallable)
        {
            return apply(connection.prepareCall(renderedQuery));
        }
        else
        {
            return apply(connection.prepareStatement(renderedQuery));
        }
    }

    public PreparedStatement render(Connection connection, String query, Pattern pattern) throws Exception
    {
        return render(connection, query, pattern, false, false);
    }
}


