/* 

Copyright 2003-2008 MicroNova (R)
All rights reserved.

Redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following
conditions are met:

    * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

    * Neither the name of MicroNova nor the names of its contributors
    may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

*/

package com.micronova.util;

import com.micronova.util.servlet.HttpDispatch;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import javax.servlet.ServletContext;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.jsp.PageContext;

public class ServletUtil
{
    public static final String COOKIE = "cookie";
    public static final String HEADER = "header";
    public static final String ERROR = "error";
    public static final String ERRORMESSAGE = "errorMessage";
    public static final String REDIRECT = "redirect";

    public static Cookie makeCookie(String name, Object object) throws Exception
    {
        if (object instanceof Cookie)
        {
            return (Cookie)object;
        }
        else if (object != null)
        {
            Cookie cookie = new Cookie(name, null);

            if (object instanceof Map)
            {
                BeanUtil.fill(cookie, (Map)object);
            }
            else
            {
                cookie.setValue(object.toString());
            }

            return cookie;
        }
        else
        {
            return null;
        }
    }

    public static void setResponse(HttpServletResponse response, Map map) throws Exception
    {
        // process cookies

        Object cookieMap = map.get(COOKIE);

        if (cookieMap instanceof Map)
        {
            Iterator iterator = ((Map)cookieMap).entrySet().iterator();

            while (iterator.hasNext())
            {
                Map.Entry entry = (Map.Entry)iterator.next();

                Object cookieName = entry.getKey();
                Object cookieValue = entry.getValue();


                if (cookieName!= null)
                {
                    Cookie cookie = makeCookie(cookieName.toString(), cookieValue);
                    
                    response.addCookie(cookie);
                }
            }
        }

        // process headers

        Object headerMap = map.get(HEADER);

        if (headerMap instanceof Map)
        {
            Iterator iterator = ((Map)headerMap).entrySet().iterator();

            while (iterator.hasNext())
            {
                Map.Entry entry = (Map.Entry)iterator.next();

                String headerKey = (String)entry.getKey();
                Object headerValue = entry.getValue();

                if (headerKey != null)
                {
                    if (headerValue instanceof String)
                    {
                        response.addHeader(headerKey, headerValue.toString());
                    }
                    else
                    {
                        List list = (List)TypeUtil.isList(headerValue);

                        if (list != null)
                        {
                            Iterator listIterator = list.iterator();
                        
                            while (listIterator.hasNext())
                            {
                                Object element = listIterator.next();
                                
                                response.addHeader(headerKey, element.toString());
                            }
                        }
                    }
                }
            }
        }

        // process error

        String error = (String)map.get(ERROR);
        String errorMessage = (String)map.get(ERRORMESSAGE);

        if (error != null)
        {
            int errorCode = Integer.parseInt(error);

            if (errorMessage == null)
            {
                response.sendError(errorCode);
            }
            else
            {
                response.sendError(errorCode, errorMessage);
            }
        }

        // process redirect

        String redirect = (String)map.get(REDIRECT);

        if (redirect != null)
        {
            response.sendRedirect(redirect);
        }
    }


    /** converts relative path (not string with a "/") to absolute, if necessary */

    public static String convertPath(String path, HttpServletRequest request)
    {
        if (!path.startsWith("/"))
        {
            path = request.getServletPath().replaceFirst("/[^/]*$", "") + "/" + path;
        }

        return path;
    }

    /** adds "/" to contextPath if not there */

    public static String convertContextPath(String contextPath)
    {
        if (contextPath != null)
        {
            if (!contextPath.startsWith("/"))
            {
                contextPath = "/" + contextPath;
            }
        }

        return contextPath;
    }

    /** call properties */

    public final static String BODYPROPERTY = "parameter";
    public final static String RETURNPROPERTY = "_return";
    public final static String VALUEPROPERTY = "value";

    /** dispatch call */

    public static Object call(ServletContext context, HttpServletRequest request, HttpServletResponse response, String path, String contextPath, NestedMap map, boolean doesForward) throws Exception
    {
        Object value = null;

        path = convertPath(path, request);
        contextPath = convertContextPath(contextPath);

        Map returnMap = map.getSubMap(RETURNPROPERTY);
        
        returnMap.put(VALUEPROPERTY, null);

        if (doesForward)
        {
            HttpDispatch.forward(context, request, response, path, contextPath, map);
        }
        else
        {
            value = HttpDispatch.include(context, request, response, path, contextPath, map);
        }
        
        Object returnValue = returnMap.get(VALUEPROPERTY);

        if ((returnValue != null) && (!("".equals(returnValue))))
        {
            value = returnValue;
        }

        return value;
    }

    /** dispatch call using PageContext */

    public static Object call(PageContext pageContext, String path, String contextPath, NestedMap map, boolean doesForward) throws Exception
    {
        return call(pageContext.getServletContext(), (HttpServletRequest)pageContext.getRequest(), (HttpServletResponse)pageContext.getResponse(), path, contextPath, map, doesForward);
    }
}

