/* 

Copyright 2003-2008 MicroNova (R)
All rights reserved.

Redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following
conditions are met:

    * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

    * Neither the name of MicroNova nor the names of its contributors
    may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

*/

package com.micronova.util;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.Attr;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import org.w3c.dom.Document;

/** XML utilities */

public class XMLMap
{
    /** control keys */

    public static final String TYPEATTRIBUTE = "typeAttribute";
    public static final String INDEXATTRIBUTE = "indexAttribute";
    public static final String DEFAULTTYPE = "defaultType";
    public static final String MAPTYPE = "mapType";
    public static final String LISTTYPE = "listType";
    public static final String ITEMNAME = "itemName";
    public static final String LISTNAME = "listName";
    public static final String NAMEPREFIX = "namePrefix";
    public static final String NAMESPACE = "namespace";
    public static final String ROOTNAME = "rootName";
    public static final String NAMEMAP = "nameMap";
    public static final String ROOTATTRIBUTE = "rootAttribute";
    public static final String HASATTRIBUTES = "hasAttributes";
    public static final String DOESENCODENAMES = "doesEncodeNames";

    public static final String ATTRIBUTE = "^";

    public static boolean isNameStartChar(char c)
    {
        // ignore ':' for encoding

        return (c == '_' || (c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z'));
    }

    public static boolean isNameChar(char c)
    {
        return (isNameStartChar(c) || c == '-' || c == '.' || (c >= '0' && c <= '9'));
    }

    public static String encodeName(String name)
    {
        char[] chars = name.toCharArray();
        int length = chars.length;

        if (length == 0)
        {
            return "_x_";
        }

        StringBuffer buffer = new StringBuffer();

        int index = 0;
        boolean afterUnderscore = false;

        if (length > 0)
        {
            char c = chars[0];

            if (c == '_')
            {
                afterUnderscore = true;
                index ++;
            }
            else if (isNameStartChar(c))
            {
                buffer.append(c);
                index ++;
            }
            else if (isNameChar(c))
            {
                buffer.append("_0");
            }
        }
       
        while (index < length)
        {
            char c = chars[index];

            if (afterUnderscore)
            {
                buffer.append('_');
                
                if ((c == '0') || (c == 'x') || (c == '_'))
                {
                    buffer.append('_');
                }
                
                afterUnderscore = false;
            }

            if (c == '_')
            {
                afterUnderscore = true;
            }
            else if (isNameChar(c))
            {
                buffer.append(c);
            }
            else
            {
                buffer.append("_x");
                String hex = Integer.toHexString((int)c);
                buffer.append(hex);
                buffer.append("_");
            }

            index ++;
        }

        if (afterUnderscore)
        {
            buffer.append("_");
        }

        return buffer.toString();
    }

    public static String decodeName(String encoded)
    {
        if (encoded.equals("_x_"))
        {
            return "";
        }

        StringBuffer buffer = new StringBuffer();
        StringBuffer hexBuffer = null;

        char[] chars = encoded.toCharArray();
        int length = chars.length;
        boolean afterUnderscore = false;

        for (int index = 0; index < length; index ++)
        {
            char c = chars[index];

            if (afterUnderscore)
            {
                if (c == '0')
                {
                    // ignore
                }
                else if (c == '_')
                {
                    buffer.append('_');
                }
                else if (c == 'x')
                {
                    if (hexBuffer == null)
                    {
                        hexBuffer = new StringBuffer();
                    }

                    hexBuffer.setLength(0);

                    for (int j = index + 1; j < length; j ++)
                    {
                        char d = chars[j];

                        if (d == '_')
                        {
                            index = j;
                            break;
                        }
                        else
                        {
                            hexBuffer.append(d);
                        }
                    }

                    buffer.append((char)Integer.parseInt(hexBuffer.toString(), 16));
                }
                else
                {
                    buffer.append('_');
                    buffer.append(c);
                }

                afterUnderscore = false;
            }
            else
            {
                if (c == '_')
                {
                    afterUnderscore = true;
                }
                else
                {
                    buffer.append(c);
                }
            }
        }

        if (afterUnderscore)
        {
            buffer.append('_');
        }

        return buffer.toString();
    }
   

    /** default control map */

    public static final Map defaultControlMap;

    static
    {
        defaultControlMap = new HashMap();
        
        defaultControlMap.put(TYPEATTRIBUTE, null);
        defaultControlMap.put(INDEXATTRIBUTE, "index");
        defaultControlMap.put(DEFAULTTYPE, "string");
        defaultControlMap.put(MAPTYPE, "map");
        defaultControlMap.put(LISTTYPE, "list");
        defaultControlMap.put(LISTNAME, "_");
        defaultControlMap.put(ITEMNAME, "item");

        defaultControlMap.put(ROOTNAME, "root");
        defaultControlMap.put(NAMESPACE, "*");
        defaultControlMap.put(DOESENCODENAMES, "true");
    }

    /** get control map */

    protected static final Map getControlMap(Map map, String key)
    {
        Map value = null;

        if (map != null)
        {
            value = (Map)map.get(key);
        }

        if (value == null)
        {
            value = (Map)defaultControlMap.get(key);
        }

        return value;
    }

    /** get control string */

    protected static final String getControlString(Map map, String key)
    {
        String value = null;

        if (map != null)
        {
            value = (String)map.get(key);
        }

        if (value == null)
        {
            value = (String)defaultControlMap.get(key);
        }

        return value;
    }

    /** add map entries as attributes */

    protected static final Element addAttribute(Element element, Map map)
    {
        if (map != null)
        {
            Iterator iterator = map.entrySet().iterator();

            while (iterator.hasNext())
            {
                Map.Entry entry = (Map.Entry)iterator.next();

                element.setAttribute(entry.getKey().toString(), entry.getValue().toString());
            }
        }

        return element;
    }

    /** generates XML representation of a map */

    public static final Document encode(Object object) throws Exception
    {
        return encode(object, null);
    }

    /** generates XML representation of a map, according to the control map */

    public static final Document encode(Object object, Map control) throws Exception
    {
        Document document = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();
        
        String name = getControlString(control, ROOTNAME);
        boolean isEmptyName = TypeUtil.isEmptyString(name);

        Element root = buildMapNode(document, (isEmptyName) ? "root" : name, object, getControlMap(control, ATTRIBUTE), control);

        if (isEmptyName)
        {
            Node firstChild = root.getFirstChild();
            if (firstChild instanceof Element)
            {
                root = (Element)firstChild;
            }
        }

        document.appendChild(root);

        return document;
    }

protected static final Element buildMapNode(Document document, String name, Object object, Map attributes, Map control)
    {
        String elementName = name;

        boolean doesEncodeNames = TypeUtil.isTrue(getControlString(control, DOESENCODENAMES));

        if (doesEncodeNames)
        {
            elementName = encodeName(elementName);
        }

        String namePrefix = getControlString(control, NAMEPREFIX);

        if (!TypeUtil.isEmpty(namePrefix))
        {
            elementName = namePrefix + elementName;
        }

        Element element = document.createElement(elementName);

        String typeAttribute = getControlString(control, TYPEATTRIBUTE);

        if (object instanceof Map)
        {
            Map elementAttributes = (Map)((Map)object).get(ATTRIBUTE);

            String mapType = getControlString(control, MAPTYPE);

            if ((!TypeUtil.isEmpty(typeAttribute)) && (!TypeUtil.isEmpty(mapType)))
            {
                element.setAttribute(typeAttribute, mapType);
            }

            TreeMap treeMap = new TreeMap((Map)object);

            Iterator iterator = treeMap.entrySet().iterator();

            while (iterator.hasNext())
            {
                Map.Entry entry = (Map.Entry)iterator.next();

                Object key = entry.getKey();
                Object value = entry.getValue();

                if (key != null)
                {
                    String keyString = key.toString();

                    if (!ATTRIBUTE.equals(keyString))
                    {
                        if (value instanceof List)
                        {
                            String itemName = getControlString(control, LISTNAME);
                            
                            if ("*".equals(itemName))
                            {
                                keyString = name;
                            }
                            else
                            {
                                keyString = itemName;
                            }
                        }

                        Map elementAttribute = null;

                        if (elementAttributes != null)
                        {
                            elementAttribute = (Map)elementAttributes.get(keyString);
                        }
                        
                        Element child = buildMapNode(document, keyString, value, elementAttribute, control);

                        element.appendChild(child);
                    }
                }
            }
        }
        else if (object instanceof List)
        {
            String listType = getControlString(control, LISTTYPE);
            String indexAttribute = getControlString(control, INDEXATTRIBUTE);

            if ((!TypeUtil.isEmpty(typeAttribute)) && (!TypeUtil.isEmpty(listType)))
            {
                element.setAttribute(typeAttribute, listType);
            }

            int index = 0;

            Iterator iterator = ((List)object).iterator();

            String itemName = getControlString(control, ITEMNAME);

            while (iterator.hasNext())
            {
                Object item = iterator.next();

                Element child = buildMapNode(document, itemName, item, attributes, control);

                if (!TypeUtil.isEmpty(indexAttribute))
                {
                    child.setAttribute(indexAttribute, Integer.toString(index));
                }

                element.appendChild(child);

                index ++;
            }
        }
        else 
        {
            if (object != null)
            {
                String defaultType = getControlString(control, DEFAULTTYPE);
                
                if ((!TypeUtil.isEmpty(typeAttribute)) && (!TypeUtil.isEmpty(defaultType)))
                {
                    element.setAttribute(typeAttribute, defaultType);
                }

                Text text = document.createTextNode(object.toString());
                element.appendChild(text);
            }
        }

        addAttribute(element, attributes);

        return element;
    }

    /** convert XML back to NestedMap, using null namespace */

    public static Object decode(Node node) throws Exception
    {
        return decode(node, null);
    }

    /** get node name, according to "namespace" */

    public static String getNodeName(Node node, String namespace)
    {
        String nodeName = null;

        String namespaceURI = node.getNamespaceURI();

        String localName = node.getLocalName();

        if (localName != null)
        {
            nodeName = localName;
        }
        else
        {
            nodeName = node.getNodeName();
        }

        if (!"_".equals(nodeName))
        {
            if (!"*".equals(namespace))
            {
                if (namespaceURI != null)
                {
                    if (!namespaceURI.equals(namespace))
                    {
                        nodeName = namespaceURI + "|" + nodeName;
                    }
                }
                else
                {
                    if ((namespace != null) && (!namespace.equals("")))
                    {
                        nodeName = "|" + nodeName;
                    }
                }
            }
        }

        return nodeName;
    }

    /** get attribute map */

    public static NestedMap getAttributeMap(Node node, String namespace) throws Exception
    {
        NestedMap attributeMap = null;

        NamedNodeMap attributeNodeMap = node.getAttributes();
        
        if (attributeNodeMap != null)
        {
            int attributeCount = attributeNodeMap.getLength();
                        
            if (attributeCount > 0)
            {
                attributeMap = new NestedMap();
                
                while (--attributeCount >= 0)
                {
                    Attr attr = (Attr)attributeNodeMap.item(attributeCount);
                                
                    if (attr != null)
                    {
                        String attributeName = getNodeName(attr, namespace);
                        
                        
                        attributeMap.put(attributeName, attr.getValue());
                    }
                }
            }
        }

        return attributeMap;
    }

    /** convert XML back to NestedMap */

    public static Object decode(Node node, Map control) throws Exception
    {
        return decode(node, control, "/");
    }

    /** convert XML back to NestedMap */

    public static Object decode(Node node, Map control, String path) throws Exception
    {
        NestedMap map = null;

        String text = null;

        node.normalize();

        NodeList children = node.getChildNodes();

        int childrenLength = children.getLength();

        String namespace = getControlString(control, NAMESPACE);
        boolean hasAttributes = TypeUtil.isTrue(getControlString(control, HASATTRIBUTES));
        boolean doesEncodeNames = TypeUtil.isTrue(getControlString(control, DOESENCODENAMES));

        Map nameMap = getControlMap(control, NAMEMAP);

        for (int i = 0; i < childrenLength; i ++)
        {
            Node child = children.item(i);

            int nodeType = child.getNodeType();

            if (nodeType == Node.ELEMENT_NODE)
            {
                if (map == null)
                {
                    map = new NestedMap();
                }

                String nodeName = getNodeName(child, namespace);

                NestedMap attributeMap = null;
                
                if (hasAttributes)
                {
                    NamedNodeMap attributeNodeMap = child.getAttributes();

                    if (attributeNodeMap != null)
                    {
                        attributeMap = getAttributeMap(child, namespace);
                    }
                }

                String nextPath = path + nodeName + "/";

                String target = getControlString(nameMap, nextPath);

                if (target == null)
                {
                    target = getControlString(nameMap, nodeName);
                    
                    if (target == null)
                    {
                        target = nodeName;
                    }
                }

                if (doesEncodeNames)
                {
                    target = decodeName(target);
                }

                if ("_".equals(target))
                {
                    NodeList listItems = child.getChildNodes();
                    int listLength = listItems.getLength();

                    for (int j = 0; j < listLength; j ++)
                    {
                        Node listItem = listItems.item(j);

                        NestedMap listItemAttributeMap = getAttributeMap(listItem, namespace);

                        String indexProperty = "*";

                        if (listItemAttributeMap != null)
                        {
                            indexProperty = listItemAttributeMap.getString("index", indexProperty);
                        }

                        Object decoded = decode(listItem, control, nextPath);

                        if (decoded != null)
                        {
                            map.put("@_." + indexProperty, decoded);

                            if (hasAttributes && (listItemAttributeMap != null))
                            {
                                map.put("@" + ATTRIBUTE + "._." + indexProperty, listItemAttributeMap);
                            }
                        }
                    }
                }
                else 
                {
                    map.put(target, decode(child, control, nextPath));

                    if (attributeMap != null)
                    {
                        map.put("@" + ATTRIBUTE + "." + target, attributeMap);
                    }
                }
            }
            else if (nodeType == Node.TEXT_NODE)
            {
                if (text == null)
                {
                    text = child.getNodeValue();
                }
                else
                {
                    text = text + child.getNodeValue();
                }
            }
        }

        if (map != null)
        {
            return map;
        }
        else
        {
            return text;
        }
    }

    public static Object decode(String string, Map control) throws Exception
    {
        return decode(XMLUtil.parse(string, control), control, "/");
    }

    public static Object decode(String string) throws Exception
    {
        return decode(string, null);
    }
}
