/* 

Copyright 2003-2008 MicroNova (R)
All rights reserved.

Redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following
conditions are met:

    * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

    * Neither the name of MicroNova nor the names of its contributors
    may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

*/


package com.micronova.util;

import java.io.InputStream;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

/** XML utilities */

public class XMLUtil
{
    /** tag encoding/decoding maps */

    private static final Pattern patternEncode;
    private static final Map replacementMapEncode;

    private static final Pattern patternDecode;
    private static final Map replacementMapDecode;

    static
    {
        patternDecode = Pattern.compile("(&amp;|&quot;|&lt;|&gt;|&apos;|&#034;|&#039;|&#123;|&#125;)");

        replacementMapDecode = new HashMap();

        replacementMapDecode.put("&amp;", "&");
        replacementMapDecode.put("&quot;", "\"");
        replacementMapDecode.put("&#034;", "\"");
        replacementMapDecode.put("&apos;", "'");
        replacementMapDecode.put("&#039;", "'");
        replacementMapDecode.put("&lt;", "<");
        replacementMapDecode.put("&gt;", ">");
        replacementMapDecode.put("&#123;", "{");
        replacementMapDecode.put("&#125;", "}");

        patternEncode = Pattern.compile("[&\"'<>{}]");

        replacementMapEncode = new HashMap();

        replacementMapEncode.put("&", "&amp;");
        replacementMapEncode.put("\"", "&quot;");
        replacementMapEncode.put("'", "&#039;");
        replacementMapEncode.put("<", "&lt;");
        replacementMapEncode.put(">", "&gt;");
        replacementMapEncode.put("{", "&#123;");
        replacementMapEncode.put("}", "&#125;");
    }

    /** decode tag escapes */

    public static String decode(String string)
    {
        return StringUtil.applyPattern(string, patternDecode, replacementMapDecode);
    }

    /** encode tag characters */

    public static String encode(String string)
    {
        return encode(string, null);
    }

    /** encode tag characters, with specified pattern */

    public static String encode(String string, Pattern pattern)
    {
        if (pattern == null)
        {
            pattern = patternEncode;
        }

        return StringUtil.applyPattern(string, pattern, replacementMapEncode);
    }

    /** create a Source from an Object */

    public static final Source createSource(Object object) throws Exception
    {
        Source source = null;

        if (object instanceof Node)
        {
            source = new DOMSource((Node)object);
        }
        else if (object instanceof String)
        {
            source = new StreamSource(new StringReader(object.toString()));
        }
        else if (object instanceof InputSource)
        {
            source = new SAXSource((InputSource)object);
        }
        else if (object instanceof InputStream)
        {
            source = new StreamSource((InputStream)object);
        }
        else if (object instanceof Reader)
        {
            source = new StreamSource((Reader)object);
        }
        else if (object instanceof URL)
        {
            source = new StreamSource(((URL)object).toString());
        }
        
        return source;
    }

    /** outputs Source */

    public static final String output(Source source, NestedMap controlMap) throws Exception
    {
        if (source != null)
        {
            Transformer transformer = TransformerFactory.newInstance().newTransformer();
        
            if (controlMap != null)
            {
                Iterator iterator = controlMap.entrySet().iterator();

                while (iterator.hasNext())
                {
                    Map.Entry entry = (Map.Entry)iterator.next();

                    transformer.setOutputProperty(entry.getKey().toString(), entry.getValue().toString());
                }
            }
            
            StringWriter writer = new StringWriter();
            Result result = new StreamResult(writer);
            
            transformer.transform(source, result);

            return writer.toString();
        }
        else
        {
            return null;
        }
    }

    /** parses document according to control map */

    public static final Document parse(Object object, Map controlMap) throws Exception
    {
        Document document = null;

        if (object != null)
        {
            DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();

            if (controlMap != null)
            {
                BeanUtil.fill(documentBuilderFactory, controlMap);
            }
            else
            {
                documentBuilderFactory.setNamespaceAware(true);
                documentBuilderFactory.setIgnoringComments(true);
                documentBuilderFactory.setIgnoringElementContentWhitespace(true);
                documentBuilderFactory.setExpandEntityReferences(true);
                documentBuilderFactory.setCoalescing(true);
                documentBuilderFactory.setValidating(false);
            }

            DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
                
            Reader reader;

            if (object instanceof Reader)
            {
                reader = (Reader)object;
            }
            else
            {
                reader = new StringReader(object.toString());
            }

            document = documentBuilder.parse(new InputSource(reader));
        }

        return document;
    }

    private static Map getParseHtmlMap(Object object)
    {
        Map map = null;

        if (object != null)
        {
            if (object instanceof Map)
            {
                map = (Map)object;
            }
            else
            {
                List list = TypeUtil.isStringList(object, ',', '\\');

                if ((list != null) && (list.size() > 0))
                {
                    map = new HashMap();
                    
                    Iterator iterator = list.iterator();
                    
                    while (iterator.hasNext())
                    {
                        map.put(iterator.next().toString().toLowerCase(), Boolean.TRUE);
                    }
                }
            }
        }

        return map;
    }

    private static int getParseHtmlStrategy(String strategyName)
    {
        int strategy = com.micronova.util.cc.html.Parser.STRATEGY_SINGLE;

        if ("data".equals(strategyName))
        {
            strategy = com.micronova.util.cc.html.Parser.STRATEGY_DATA;
        }
        else if ("none".equals(strategyName))
        {
            strategy = com.micronova.util.cc.html.Parser.STRATEGY_NONE;
        }
        else if ("single".equals(strategyName))
        {
            strategy = com.micronova.util.cc.html.Parser.STRATEGY_SINGLE;
        }

        return strategy;
    }

    public static Document parseHtml(Object object, NestedMap controlMap) throws Exception
    {
        Document document = null;
        Reader reader = null;

        try
        {
            String prefix = controlMap.getString("prefix", "");
            int strategy = getParseHtmlStrategy(controlMap.getString("strategy"));
            String rootName = controlMap.getString("rootName", "root");
            Map exclude = getParseHtmlMap(controlMap.get("exclude"));
            Map include = getParseHtmlMap(controlMap.get("include"));

            if (object instanceof Reader)
            {
                reader = (Reader)object;
            }
            else
            {
                reader = new StringReader(object.toString());
            }

            com.micronova.util.cc.html.Parser parser = new com.micronova.util.cc.html.Parser(reader);
          
            document = parser.buildDocument(prefix, strategy, rootName, exclude, include);
        }
        catch (Exception e)
        {
            try
            {
                reader.close();
            }
            catch (Exception ee)
            {
            }

            throw e;
        }

        return document;
    }

    public static final String[] NODETYPE =
    {
        null,
        "Element",
        "Attr",
        "Text",
        "CDATASection",
        "EntityReference",
        "Entity",
        "ProcessingInstruction",
        "Comment",
        "Document",
        "DocumentType",
        "DocumentFragment",
        "Notation"
    };

    public static final Object encodeNodeMap(Node node)
    {
        if (node != null)
        {
            short type = node.getNodeType();
           
            if (type == Node.TEXT_NODE)
            {
                return node.getNodeValue();
            }
            else
            {
                NestedMap map = new NestedMap();
                map.put("type", NODETYPE[node.getNodeType()]);
                map.put("name", node.getNodeName());
                map.put("value", node.getNodeValue());

                if (node.hasAttributes())
                {
                    NamedNodeMap attributes = node.getAttributes();

                    NestedMap attributeMap = new NestedMap();

                    int length = attributes.getLength();
                    
                    for (int i = 0; i < length; i ++)
                    {
                        Attr attr = (Attr)attributes.item(i);

                        attributeMap.put(attr.getName(), attr.getValue());
                    }
                    
                    map.put("attributes", attributeMap);
                }

                if (node.hasChildNodes())
                {
                    NodeList children = node.getChildNodes();
                    int length = children.getLength();
                
                    List list = map.getSubList();
                    
                    for (int i = 0; i < length; i ++)
                    {
                        Object child = encodeNodeMap(children.item(i));
                        list.add(child);
                    }
                }

                return map;
            }
        }

        return null;
    }

    protected static final Node _decodeNodeMap(Document document, Object object)
    {
        Node node = null;

        if (object != null)
        {
            if (object instanceof Node)
            {
                node = (Node)object;
            }
            else if (object instanceof String)
            {
                node = document.createTextNode(object.toString());
            }
            else if (object instanceof Map)
            {
                NestedMap map = TypeUtil.isNestedMap(object);

                String type = map.getString("type", "Element");

                if ("Element".equals(type))
                {
                    Element element = document.createElement(map.getString("name"));
                    Map attributes = map.getSubMap("attributes");
                    
                    Iterator iterator = attributes.entrySet().iterator();
                    
                    while (iterator.hasNext())
                    {
                        Map.Entry entry = (Map.Entry)iterator.next();
                        element.setAttribute(entry.getKey().toString(), entry.getValue().toString());
                    }

                    node = element;
                }
                else if ("Text".equals(type))
                {
                    node = document.createTextNode(map.getString("value"));
                }
                else if ("CDATASection".equals(type))
                {
                    node = document.createCDATASection(map.getString("value"));
                }
                else if ("Comment".equals(type))
                {
                    node = document.createComment(map.getString("value"));
                }
                else if ("Document".equals(type))
                {
                    node = document;
                }
                else if ("DocumentFragment".equals(type))
                {
                    node = document.createDocumentFragment();
                }
                else if ("EntityReference".equals(type))
                {
                    node = document.createEntityReference(map.getString("name"));
                }
                else if ("ProcessingInstruction".equals(type))
                {
                    node = document.createProcessingInstruction(map.getString("name"), map.getString("value"));
                }
                else if ("Attr".equals(type))
                {
                    Attr attr = document.createAttribute(map.getString("name"));
                    attr.setValue(map.getString("value"));

                    node = attr;
                }

                if (node != null)
                {
                    List children = map.getSubList();
                
                    Iterator iterator = children.iterator();
                    
                    while (iterator.hasNext())
                    {
                        node.appendChild(_decodeNodeMap(document, iterator.next()));
                    }
                }
            }
        }

        return node;
    }

    public static final Node decodeNodeMap(Object object, Object controlSpec) throws Exception
    {
        Document document = null;

        if (controlSpec instanceof Document)
        {
            document = (Document)controlSpec;
        }
        else
        {
            DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();

            if (controlSpec != null)
            {
                BeanUtil.fill(documentBuilderFactory, TypeUtil.isNestedMap(controlSpec));
            }

            DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
            document = documentBuilder.newDocument();
        }

        return _decodeNodeMap(document, object);
    }

    public static final Node decodeNodeMap(Object object) throws Exception
    {
        return decodeNodeMap(object, null);
    }

    /** encode NestedMap into XML attribute list (with starting space) */

    public static final StringBuffer encodeAttrList(StringBuffer buffer, NestedMap map)
    {
        Iterator iterator = map.entrySet().iterator();
        
        while (iterator.hasNext())
        {
            Map.Entry entry = (Map.Entry)iterator.next();

            String keyString = entry.getKey().toString();

            if (!(NestedMap.LISTKEY.equals(keyString)))
            {
                buffer.append(" ");
                buffer.append(XMLMap.encodeName(keyString));
                buffer.append("=\"");
                
                Object value = entry.getValue();

                if (value instanceof Map)
                {
                    value = StringUtil.encodeCSS((Map)value);
                }
                
                buffer.append(encode(value.toString()));
                buffer.append("\"");
            }
        }

        return buffer;
    }

    public static final String encodeAttrList(NestedMap map)
    {
        StringBuffer buffer = new StringBuffer();
        buffer = encodeAttrList(buffer, map);
        return buffer.toString();
    }
}
