/* 

Copyright 2003-2008 MicroNova (R)
All rights reserved.

Redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following
conditions are met:

    * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

    * Neither the name of MicroNova nor the names of its contributors
    may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

*/

package com.micronova.util.codec;

import com.micronova.util.NestedMap;
import com.micronova.util.TypeUtil;
import com.micronova.util.XMLUtil;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/** XML-related codecs */

public class CodecXML extends Codec
{
    /** parses document */

    public static final Object parse(Object object, Object control) throws Exception
    {
        return XMLUtil.parse(object, TypeUtil.isNestedMap(control));
    }

    /** parser with default controlMap */

    public static Object parse(Object object) throws Exception
    {
        return parse(object, null);
    }

    /** encodes XML tag characters */

    public static final Object encode(Object object) throws Exception
    {
        return encode(object, null);
    }

    /** encodes XML tag characters specified by the "spec" regex (e.g., "[&<>]" to encode &, <, > only); if "spec" is null, then all supported characters are encoded.  If "spec" is "~", then "[&<>]" is assumed. */

    public static final Object encode(Object object, Object spec) throws Exception
    {
        if (object != null)
        {
            if ("~".equals(spec))
            {
                spec = "[&<>]";
            }

            object = XMLUtil.encode(object.toString(), TypeUtil.isPattern(spec));
        }

        return object;
    }

    /** decodes XML tag characters */

    public static final Object decode(Object object) throws Exception
    {
        if (object != null)
        {
            object = XMLUtil.decode(object.toString());
        }

        return object;
    }

    /** outputs an XML object */

    public static final Object output(Object object, Object control) throws Exception
    {
        return XMLUtil.output(XMLUtil.createSource(object), new NestedMap(control));
    }

    /** outputs using default control (omit-xml-declaration=yes) */

    public static final Object output(Object object) throws Exception
    {
        return output(object, "omit-xml-declaration=yes");
    }

    /** parses HTML as in parseHtml */

    public static final Object parseHtml(Object object, Object control) throws Exception
    {
        if (object != null)
        {
            object = XMLUtil.parseHtml(object, TypeUtil.isNestedMap(control));
        }
        
        return object;
    }

    public static final Object parseHtml(Object object) throws Exception
    {
        return parseHtml(object, null);
    }

    /** removes a node from its parent */

    public static final Object removeNode(Object object) throws Exception
    {
        Node node = TypeUtil.isNode(object);

        if (node != null)
        {
            Node parentNode = node.getParentNode();

            if (parentNode != null)
            {
                parentNode.removeChild(node);
            }
        }
        
        return object;
    }

    /** insert a node before the second argument. */

    public static final Object insertNodeBefore(Object object, Object before) throws Exception
    {
        Node beforeNode = TypeUtil.isNode(before);

        if (beforeNode != null)
        {
            Node node = TypeUtil.isNode(object, beforeNode.getOwnerDocument());

            if (node != null)
            {
                Node parentNode = beforeNode.getParentNode();

                if (parentNode != null)
                {
                    parentNode.insertBefore(node, beforeNode);
                }
            }
        }
        
        return object;
    }

    /** appends a child */

    public static final Object appendNode(Object object, Object parent) throws Exception
    {
        Node parentNode = TypeUtil.isNode(parent);

        if (parentNode != null)
        {
            Node node = TypeUtil.isNode(object, parentNode.getOwnerDocument());

            if (node != null)
            {
                parentNode.appendChild(node);
            }
        }
        
        return object;
    }

    /** shallow copies a node */

    public static final Object cloneNode(Object object) throws Exception
    {
        Node node = TypeUtil.isNode(object);

        if (node != null)
        {
            node = node.cloneNode(false);
        }

        return node;
    }
    
    /** deep copies a node */

    public static final Object copyNode(Object object) throws Exception
    {
        Node node = TypeUtil.isNode(object);

        if (node != null)
        {
            node = node.cloneNode(true);
        }

        return node;
    }

    /** replace an old node */

    public static final Object replaceNode(Object object, Object old) throws Exception
    {
        Node node = null;
        Node oldNode = TypeUtil.isNode(old);

        if (oldNode != null)
        {
            node = TypeUtil.isNode(object, oldNode.getOwnerDocument());

            if (node != null)
            {
                Node parentNode = oldNode.getParentNode();

                if (parentNode != null)
                {
                    parentNode.replaceChild(node, oldNode);
                }
            }
        }

        return node;
    }

    /** normalizes a node */

    public static final Object normalizeNode(Object object) throws Exception
    {
        Node node = TypeUtil.isNode(object);

        if (node != null)
        {
            node.normalize();
        }

        return node;
    }

    /** sets element attribute; if value is null, then attribute is removed. */

    public static final Object setAttribute(Object object, Object key, Object value) throws Exception
    {
        Node node = TypeUtil.isNode(object);

        if (node instanceof Element)
        {
            if (key != null)
            {
                String keyString = key.toString();

                if (value == null)
                {
                    ((Element)node).removeAttribute(keyString);
                }
                else
                {
                    ((Element)node).setAttribute(keyString, value.toString());
                }
            }
        }

        return object;
    }

    /** gets element attribute value */

    public static final Object getAttribute(Object object, Object key) throws Exception
    {
        Node node = TypeUtil.isNode(object);

        if (node instanceof Element)
        {
            if (key != null)
            {
                object = ((Element)node).getAttribute(key.toString());
            }
        }

        return object;
    }
}

